/* gafstoken.c 

	A generic routiue to get an AFS token by calling
	an external program. 

	This routine is generic in that it does not depend on 
	any specific implementation of AFS, or Kerberos.

	It can be built on a system without AFS, and if run
	on a system without AFS it will catch the signals
	and continue. 

	This will allow a deamon to be configured, compiled
	and linked without any AFS or Kerberos dependencies
	and can thus avoid conflicts between these packages.
	The resulting deamon executable can be run on a system
	with or without AFS.  
	
	Examples of external programs that could be called
	are aklog, afslog, gssklog, klog, ak5log or even a script
	to call one of these. 

	The external program is expected to support 
	the -p <homedir> option. 

	Parameters to gafstoken: 

		char * external_program  - path to the executable
			this would be set by configure, or defaulted
			in some way. If NULL a built in location is used.
							
		char **env - Environment to run under. This
			should include any environemt variables needed
			by the external program such as KRB5CCNAME
			or X509_USER_PROXY These would have been set
			by GSSAPI or PAM.  If NULL, the current environment
			is used. 
	
		uid_t uid - the uid of the user  that the external_program
			is run under.

		char * homedir - A file location, if in AFS the 
			external program will attempt to get a token
			for the cell containing it.  If NULL atempt
			to get a token for the default cell of the client.

		int setpag - Used to get a PAG using syscall 
			We no longer try and let the external program 
			try and set the PAG of the parent. 

	Returns:
		-1 - some system error, see errno;
		 0 - the external program if present was called and 
			 MAY have worked. 

	The caller should not need to test the return code, and should
	continue with or without the AFS token or PAG. 

	*/

#include <errno.h>
#ifdef HAVE_WAIT_H
#include <wait.h>
#endif
#include <sys/stat.h> 
#include <fcntl.h>
#include <unistd.h>

#ifdef DEBUG
#include <stdio.h>
#endif

/* FOR NOW ALL HAVE POSIX_SIGNALS */
#define POSIX_SIGNALS
#ifdef POSIX_SIGNALS 
#define POSIX_SETJMP
#endif
#define HAVE_WAITPID
#define WAIT_USES_INT

#include <signal.h>
#include <setjmp.h>

#include "gafstoken.h"

#ifndef _PATH_AFS_EXTERNAL_PROGRAM
#define _PATH_AFS_EXTERNAL_PROGRAM "/krb5/libexec"
#endif 

#ifndef _AFS_EXTERNAL_PROGRAM
#define _AFS_EXTERNAL_PROGRAM "get-afs-token"
#endif 

	

/* 
 * We need the AFS_SYSCALL and AFSCALL_SETPAG numbers, but we
 * want to make this independent of AFS so it can be built
 * and used on systems which don't have AFS. 
 *
 * Let me say that again, you can build on a system
 * that does not have AFS! If run on a system without
 * AFS we will trap the failure, and continue. 
 *
 * See:
 * #include "afs/param.h"
 * #include "afs/afs.h"
 */

#ifdef __linux
#include <sys/ioctl.h>
#define PROC_SYSCALL_FNAME "/proc/fs/openafs/afs_ioctl"
#define PROC_SYSCALL_ARLA_FNAME "/proc/fs/nnpfs/afs_ioctl"
#define VIOC_SYSCALL_TYPE 'C' 
#define VIOC_SYSCALL _IOW(VIOC_SYSCALL_TYPE,1,void *)

struct afsprocdata {
  long param4;
  long param3;
  long param2;
  long param1;
  long syscall;
};
static
int proc_afs_syscall(long syscall, long param1, long param2, long param3,
             long param4, int *rval);
#endif

#define AFSCALL_SETPAG 21
 
#if defined(sun) && defined(__SVR4)
#include <sys/systeminfo.h>
/* Solaris  5.6 and below uses 105 */
/* But Solaris 5.7 with ASF 3.5 uses 73 */
/* and Solaris 5.8 and 5.9 use 65 */
/* we will determing the number at runtime below. */

#ifndef AFS_SYSCALL
#define AFS_SYSCALL 65 
#endif

#elif defined(sun)
/* SunOS uses 31 */
#define AFS_SYSCALL 31

#elif defined(sgi) || defined(_sgi)
/* sgi does it different, each AFSCALL has its own syscall!
 * so define the SETPAG one
 */
#define AFS_SYSCALL 65+1000
#undef AFSCALL_SETPAG 
#define AFSCALL_SETPAG 0

#elif defined(hpux) || defined(_hpux) ||defined(__hpux)
#define AFS_SYSCALL 48

#elif defined(alpha)
#define AFS_SYSCALL 232

#elif defined(linux)
#define AFS_SYSCALL 137

#elif defined(_AIX)
#define AFS_SYSCALL -1
/* AFS_SYSCALL not really used for AIX */

#else
#define AFS_SYSCALL (Unknown_AFS_SYSCALL)
#endif

/* Only run this AFS PAG code on systems with POSIX
 * All that we are interested in do: AIX 4.x,
 * Solaris 2.x, HPUX 10.x, HPUX 11 and SGI 6.5.
 * This simplifies the build/configure which I don't 
 * want to change now.
 * They also all have waitpid  and wait uses int was well. 
 */

#define HAVE_WAITPID
#define WAIT_USES_INT

#include <signal.h>
#include <setjmp.h>
#ifndef POSIX_SETJMP
#undef sigjmp_buf
#undef sigsetjmp
#undef siglongjmp
#define sigjmp_buf  jmp_buf
#define sigsetjmp(j,s)  setjmp(j)
#define siglongjmp  longjmp
#endif

/* Linux does not have the SIGSYS */
#ifndef SIGSYS
#define SIGSYS SIGUSR2
#endif

#ifdef POSIX_SIGNALS
typedef void sigtype;
typedef struct sigaction handler;
#define handler_init(H,F)       (sigemptyset(&(H).sa_mask), \
                     (H).sa_flags=0, \
                     (H).sa_handler=(F))
#define handler_swap(S,NEW,OLD)     sigaction(S, &NEW, &OLD)
#define handler_set(S,OLD)      sigaction(S, &OLD, NULL)
#else
typedef sigtype (*handler)();
#define handler_init(H,F)       ((H) = (F))
#define handler_swap(S,NEW,OLD)     ((OLD) = signal ((S), (NEW)))
#define handler_set(S,OLD)      (signal ((S), (OLD)))
#endif

#ifdef  WAIT_USES_INT
                int wait_status;
#else   /* WAIT_USES_INT */
                union wait wait_status;
#endif  /* WAIT_USES_INT */


static sigjmp_buf setpag_buf;

static sigtype mysig()
{
  siglongjmp(setpag_buf, 1);
}

/* 
 * On linux, we may not have a syscall, but use a /proc
 * file  See OpenAFS src/sys/afssyscalls.c
 */

#ifdef __linux
static
int proc_afs_syscall(long syscall, long param1, long param2, long param3, 
             long param4, int *rval) {
  struct afsprocdata syscall_data;
  int fd = open(PROC_SYSCALL_FNAME, O_RDWR);
  if(fd < 0)
      fd = open(PROC_SYSCALL_ARLA_FNAME, O_RDWR);
  if(fd < 0)
    return -1;

  syscall_data.syscall = syscall;
  syscall_data.param1 = param1;
  syscall_data.param2 = param2;
  syscall_data.param3 = param3;
  syscall_data.param4 = param4;

  *rval = ioctl(fd, VIOC_SYSCALL, &syscall_data);

  close(fd);

  return 0;
}
#endif


/* lsetpagx - do the syscall. If the kernel extentions
 * are not loaded, on most systems it will get a SIGSYS
 * failt. On AIX it can get a SIGSEGV
 */

int 
lsetpagx()
{ 
  handler sa1, osa1;
  handler sa2, osa2;
  int ret = -2;
  int rval;
  int syscall_num = AFS_SYSCALL;

#if defined(sun) && defined(__SVR4)
	{
		char buf[256];
		long buf_len;
		buf_len = 256;
		syscall_num = 105; 
		if (sysinfo(SI_RELEASE, buf, buf_len) > 0) {
		      if (!strcmp(buf,"5.6"))  syscall_num = 105;
		 else if (!strcmp(buf,"5.7"))  syscall_num = 73;
		 else if (!strcmp(buf,"5.8"))  syscall_num = 65;
		 else if (!strcmp(buf,"5.9"))  syscall_num = 65;
		 else syscall_num = 65; /* assume newer version are also 65 */
		}
	}
#endif

  handler_init (sa1, mysig);
  handler_init (sa2, mysig);
  handler_swap (SIGSYS, sa1, osa1);
  handler_swap (SIGSEGV, sa2, osa2);

  if (sigsetjmp(setpag_buf, 1) == 0) {

#if defined(_AIX)
    static int (*apagaix)() = 0;

    if (!apagaix)
      apagaix = load(_PATH_AFS_EXTERNAL_PROGRAM "/apagaix",0,0);
    if (apagaix)
      ret = (*apagaix)();
#elif defined(__linux)
	rval = proc_afs_syscall(AFSCALL_SETPAG,0,0,0,0,&ret);
	if(rval)
		ret =  syscall(syscall_num, AFSCALL_SETPAG, 0, 0, 0, 0, 0);
#else
    ret =  syscall(syscall_num, AFSCALL_SETPAG, 0, 0, 0, 0, 0);
#endif

    handler_set (SIGSYS, osa1);
    handler_set (SIGSEGV, osa2);
    return(ret);
  }

  /* syscall failed! return -2 */
  handler_set (SIGSYS, osa1);
  handler_set (SIGSEGV, osa2);
  return(-2);
}

int
get_afs_token(char * external_program,
			  char ** env,
			  uid_t uid,
			  char * homedir,
			  int setpag)
{
	pid_t pid;
	int status;
	struct stat stx;
	char * ppath = _PATH_AFS_EXTERNAL_PROGRAM "/" _AFS_EXTERNAL_PROGRAM;
	char * args[4];  /* arg0, -p, homedir, NULL */
	int  argi = 0; 
	
#ifdef DEBUG
	FILE * stddeb= NULL;
	stddeb = fopen("/tmp/get_afs_token.log","a");
	setbuf(stddeb,NULL);
	fprintf(stddeb,"get_afs_token:%d pid=%d uid=%d euid=%d\n",
			__LINE__,getpid(),getuid(),geteuid());
	fprintf(stddeb,"   pgm=%s uid=%ld home=%s setpag=%d\n",
			external_program?external_program:"(default)",
			uid,
			homedir?homedir:"(NULL)",
			setpag); 
	
#endif

	if (setpag)   /* try to get a PAG using syscall */
		lsetpagx(); 

	if (external_program)
		ppath = external_program;

	/* test if external program exists */
	if (stat(ppath, &stx))
		return 0; /* does not exist, skip getting PAG and token */
#ifdef DEBUG
	fprintf(stddeb,"get_afs_token:%d\n", __LINE__);
#endif

	args[argi++] = _AFS_EXTERNAL_PROGRAM;
	if (homedir) {
		args[argi++] = "-p";
		args[argi++] = homedir;
	}
	args[argi] = NULL;

	if ((pid = fork()) < 0) {
		return -1; /* see errno */
	}

#ifdef DEBUG
	fprintf(stddeb,"get_afs_token:%d\n", __LINE__);
#endif
	if (pid == 0) {	
		setuid(uid); /* run as user */
		/* Don't want any output confusing the deamon */
		close(1);
		open("/dev/null",O_WRONLY);
		close(2);
		open("/dev/null",O_WRONLY);

#ifdef DEBUG
	fprintf(stddeb,"get_afs_token:%d\n", __LINE__);
#endif
		if (env)
			execve(ppath, args, env);
		else
			execv(ppath, args);
#ifdef DEBUG
	fprintf(stddeb,"get_afs_token:%d\n", __LINE__);
#endif
		exit(127); /* should never get here */
	}
		
#ifdef DEBUG
	fprintf(stddeb,"get_afs_token:%d\n", __LINE__);
#endif
	while (waitpid(pid, &status, 0) < 0) {
		if (errno != EINTR)
		break;
	}

	/* we could set a return code based on status,
	 * be we want to go on with or without a token */

#ifdef DEBUG
	fprintf(stddeb,"get_afs_token:%d\n", __LINE__);
#endif
	return 0;
}
